/** @file
===========================================================================
MODULE				: CCropFilter

DESCRIPTIONS		: Crop Filter for RGB24 and RGB32 Media

REVISION HISTORY	: Version 1.1


COPYRIGHT			: (c)CSIR, Meraka Institute 2007 all rights reserved

RESTRICTIONS		: The information/data/code contained within this file is 
the property of CSIR, Meraka Institute and has been 
classified as PUBLIC.
===========================================================================
*/
#pragma once

// Meraka includes
#include <DirectShow/CustomBaseFilter.h>

// {F5E69199-7AB2-40f0-BFC0-6938430F4F19}
static const GUID CLSID_CropFilter = 
{ 0xf5e69199, 0x7ab2, 0x40f0, { 0xbf, 0xc0, 0x69, 0x38, 0x43, 0xf, 0x4f, 0x19 } };

// Forward declarations
class PicCropperBase;

class CCropFilter : public CCustomBaseFilter
{
public:
	/// Constructor
	CCropFilter();
	/// Destructor
	~CCropFilter();

	/// Static object-creation method (for the class factory)
	static CUnknown * WINAPI CreateInstance(LPUNKNOWN pUnk, HRESULT *pHr); 

	/**
	* Overriding this so that we can set whether this is an RGB24 or an RGB32 Filter
	*/
	HRESULT SetMediaType(PIN_DIRECTION direction, const CMediaType *pmt);

	/**
	* Used for Media Type Negotiation 
	* Returns an HRESULT value. Possible values include those shown in the following table.
	* <table border="0" cols="2"><tr valign="top"><td><b>Value</b></td><td><b>Description</b></td></TR><TR><TD>S_OK</TD><TD>Success</TD></TR><TR><TD>VFW_S_NO_MORE_ITEMS</TD><TD>Index out of range</TD></TR><TR><TD>E_INVALIDARG</TD><TD>Index less than zero</TD></TR></TABLE>
	* The output pin's CTransformOutputPin::GetMediaType method calls this method. The derived class must implement this method. For more information, see CBasePin::GetMediaType.
	* To use custom media types, the media type can be manipulated in this method.
	*/
	HRESULT GetMediaType(int iPosition, CMediaType *pMediaType);

	/// Buffer Allocation
	/**
	* The output pin's CTransformOutputPin::DecideBufferSize method calls this method. The derived class must implement this method. For more information, see CBaseOutputPin::DecideBufferSize. 
	* @param pAlloc Pointer to the IMemAllocator interface on the output pin's allocator.
	* @param pProp Pointer to an ALLOCATOR_PROPERTIES structure that contains buffer requirements from the downstream input pin.
	* @return Value: Returns S_OK or another HRESULT value.
	*/
	HRESULT DecideBufferSize(IMemAllocator *pAlloc, ALLOCATOR_PROPERTIES *pProp);

	/**
	* The CheckTransform method checks whether an input media type is compatible with an output media type.
	* <table border="0" cols="2"> <tr valign="top"> <td  width="50%"><b>Value</b></td> <td width="50%"><b>Description</b></td> </tr> <tr valign="top"> <td width="50%">S_OK</td> <td width="50%">The media types are compatible.</td> </tr> <tr valign="top"> <td width="50%">VFW_E_TYPE_NOT_ACCEPTED</td> <td width="50%">The media types are not compatible.</td> </tr> </table>
	*/
	HRESULT CheckTransform(const CMediaType *mtIn, const CMediaType *mtOut);

	/// Overridden from CSettingsInterface
	virtual void initParameters()
	{
		addParameter(TARGET_WIDTH, &m_nOutWidth, -1, true);
		addParameter(TARGET_HEIGHT, &m_nOutHeight, -1, true);
		addParameter(LEFT_CROP, &m_nLeftCrop, -1);
		addParameter(RIGHT_CROP, &m_nRightCrop, -1);
		addParameter(TOP_CROP, &m_nTopCrop, -1);
		addParameter(BOTTOM_CROP, &m_nBottomCrop, -1);
	}

	STDMETHODIMP SetParameter( const char* type, const char* value);

	/// Overridden from CCustomBaseFilter
	virtual void InitialiseInputTypes();

private:
	/**
	* This method converts the input buffer from RGB24 | 32 to YUV420P
	* @param pSource The source buffer
	* @param pDest The destination buffer
	*/
	DWORD ApplyTransform(BYTE* pBufferIn, BYTE* pBufferOut);

	/// Sets crop if values are valid
	HRESULT SetCropIfValid(int nTotalDimensionImage, int nNewCrop, int& nOldCrop, int nOppositeCrop  );
	/// This method calculates how much should be cropped from the picture based on the crop ratios
	void RecalculateFilterParameters();

	/// Pointer to our picture cropper
	PicCropperBase* m_pCropper;

	int m_nBytesPerPixel;
	//Member variables which affect the transform
	int m_nLeftCrop;
	int m_nRightCrop;
	int m_nTopCrop;
	int m_nBottomCrop;

	int m_nStride;
	int m_nPadding;
};