#include "RtspClientSession.h"

// RTVC
#include "RtspSourceFilter.h"
#include "RtspSourceOutputPin.h"

// RTVC Live integration
#include "LiveDirectShowSink.h"
#include "H263LiveDirectShowSink.h"
#include "WavLiveDirectShowSink.h"

#include "RtspSession.h"

#include <DirectShow/GlobalDefinitions.h>
#include <Shared/DirectoryInfo.h>
#include <Shared/StringUtil.h>
#include <LiveIntegration/LiveMediaHelper.h>

// LiveMedia
#include <liveMedia.hh>
#include <BasicUsageEnvironment.hh>
#include <GroupsockHelper.hh>
#include <RTSPClient.hh>


RtspClientSession::RtspClientSession( RtspSourceFilter* pSourceFilter, bool bStreamUsingTcp)
	:m_pSourceFilter(pSourceFilter),
	env(NULL),
	m_pRtspClient(NULL),
	m_pSession(NULL),
	m_bStreamUsingTCP(bStreamUsingTcp),
	qosRecordHead(NULL),
	qosMeasurementIntervalMS(1000),
	qosMeasurementTimerTask(NULL),
	m_nTimeoutS(-1) // 10 seconds timeout
{
	
}

bool RtspClientSession::start(std::string sUrl)
{
	// Store Url
	m_sUrl = sUrl;
	const char* szUrl = m_sUrl.c_str();

	// Setup environment
	TaskScheduler* scheduler = BasicTaskScheduler::createNew();
	env = BasicUsageEnvironment::createNew(*scheduler);

	// Create client
	m_pRtspClient = RTSPClient::createNew(*env);

	bool bSuccess = false;
	do 
	{
		if (!initialiseRtspSession(szUrl))
		{
			break;
		}

		if (!afterSessionInit())
		{
			break;
		}

		if (needToPlayStream())
		{
			bSuccess = setupAndPlayRtspSession();
		}
		else
		{
			bSuccess = true;
		}
		break;
		
	} while(true);

	*env << "LiveMedia eventloop has ended\n";

	// Shutdown
	shutdown();
	
	m_pSourceFilter->ResetRtspWatchVar();
	m_pSourceFilter->SignalEndOfRtspThread();

	return bSuccess;
}

void RtspClientSession::shutdown()
{
	if (env != NULL) {
		//TODO
		/*env->taskScheduler().unscheduleDelayedTask(sessionTimerTask);
		env->taskScheduler().unscheduleDelayedTask(arrivalCheckTimerTask);
		env->taskScheduler().unscheduleDelayedTask(interPacketGapCheckTimerTask);
		env->taskScheduler().unscheduleDelayedTask(qosMeasurementTimerTask);*/
	}

	//TODO???
	/*if (qosMeasurementIntervalMS > 0) {
		printQOSData(exitCode);
	}*/

	// Close our output files:
	closeMediaSinks();

	// Teardown, then shutdown, any outstanding RTP/RTCP subsessions
	if (m_pSession && m_pRtspClient)
	{
		m_pRtspClient->teardownMediaSession(*m_pSession);
	}

	Medium::close(m_pSession);

	// Finally, shut down our client:
	if (m_pRtspClient)
		Medium::close(m_pRtspClient);
}

void RtspClientSession::closeMediaSinks()
{
	if (m_pSession == NULL) return;
	MediaSubsessionIterator iter(*m_pSession);
	MediaSubsession* subsession;
	while ((subsession = iter.next()) != NULL) 
	{
		Medium::close(subsession->sink);
		subsession->sink = NULL;
	}
}

bool RtspClientSession::createRtpSources()
{
	MediaSubsessionIterator iter(*m_pSession);
	MediaSubsession *subsession;
	Boolean madeProgress = False;
	while ((subsession = iter.next()) != NULL) {
		//if (createReceivers) 
		{
			if (!subsession->initiate()) 
			{
				*env << "Unable to create receiver for \"" << subsession->mediumName() << "/" << subsession->codecName() << "\" subsession: " << env->getResultMsg() << "\n"; 
			} 
			else 
			{
				*env << "Created receiver for \"" << subsession->mediumName() << "/" << subsession->codecName() << "\" subsession (client ports " << subsession->clientPortNum() << "-" << subsession->clientPortNum()+1 << ")\n";
				madeProgress = True;

				if (subsession->rtpSource() != NULL) 
				{
					// Because we're saving the incoming data, rather than playing
					// it in real time, allow an especially large time threshold
					// (1 second) for reordering misordered incoming packets:
					unsigned const thresh = 1000000; // 1 second 
					subsession->rtpSource()->setPacketReorderingThresholdTime(thresh);

					int socketInputBufferSize = 0;
					if (socketInputBufferSize > 0) 
					{
						// Set the RTP source's input buffer size as specified:
						int socketNum
							= subsession->rtpSource()->RTPgs()->socketNum();
						unsigned curBufferSize
							= getReceiveBufferSize(*env, socketNum);
						unsigned newBufferSize
							= setReceiveBufferTo(*env, socketNum, socketInputBufferSize);
						*env << "Changed socket receive buffer size for the \""
							<< subsession->mediumName()
							<< "/" << subsession->codecName()
							<< "\" subsession from "
							<< curBufferSize << " to "
							<< newBufferSize << " bytes\n";
					}
				}
			}
		}
	}
	return madeProgress;
}

bool RtspClientSession::setupStreams()
{
	MediaSubsessionIterator iter(*m_pSession);
	MediaSubsession *subsession;
	Boolean madeProgress = False;

	while ((subsession = iter.next()) != NULL) {
		if (subsession->clientPortNum() == 0) continue; // port # was not set

		if (!m_pRtspClient->setupMediaSubsession(*subsession, False, m_bStreamUsingTCP)) 
		{
			*env << "Failed to setup \"" << subsession->mediumName() << "/" << subsession->codecName() << "\" subsession: " << env->getResultMsg() << "\n";
		} 
		else 
		{
			*env << "Setup \"" << subsession->mediumName() << "/" << subsession->codecName()  << "\" subsession (client ports " << subsession->clientPortNum()<< "-" << subsession->clientPortNum()+1 << ")\n";
			madeProgress = True;
		}
	}
	return madeProgress;
}

bool RtspClientSession::createRtpReceivers()
{
	// Create output files:
	char* pWatchVar = m_pSourceFilter->GetRtspWatchVar();
	{
		// Create and start "FileSink"s for each subsession:
		bool madeProgress = False;
		MediaSubsessionIterator iter(*m_pSession);
		MediaSubsession *subsession;
		while ((subsession = iter.next()) != NULL) 
		{
			if (subsession->readSource() == NULL) continue; // was not initiated

			// Create an output file for each desired stream:
			//char outFileName[1000];
			//if (singleMedium == NULL) {
			//	// Output file name is
			//	//     "<filename-prefix><medium_name>-<codec_name>-<counter>"
			//	static unsigned streamCounter = 0;
			//	snprintf(outFileName, sizeof outFileName, "%s%s-%s-%d",
			//		fileNamePrefix, subsession->mediumName(),
			//		subsession->codecName(), ++streamCounter);
			//} else {
			//	sprintf(outFileName, "stdout");
			//}
			// TOREMOVE TEMP//////////////////////////////////////////////////////////////////////////
			const char* outFileName = "FileSink.dat";
			int fileSinkBufferSize = 20000;
			// TOREMOVE TEMP//////////////////////////////////////////////////////////////////////////

			FileSink* fileSink;
			//////////////////////////////////////////////////////////////////////////
			if (strcmp(subsession->mediumName(), "video") == 0 && (strcmp(subsession->codecName(), "H263-1998") == 0)) 
			{
				int nMediaCount = m_vLiveMediaSinks.size();
				LiveDirectShowSink* pSink = H263LiveDirectShowSink::createNew(nMediaCount, *env, fileSinkBufferSize, isMediaTypeKnown(), pWatchVar);
				pSink->setMediaQueue(m_pSourceFilter->GetOutputPin(nMediaCount)/*m_vOutputPins[nMediaCount]*/);
				subsession->sink = pSink;
				m_vLiveMediaSinks.push_back(pSink);
			} 
			else if (strcmp(subsession->mediumName(), "audio") == 0 && 
				((strcmp(subsession->codecName(), "L8") == 0)) ||
				(strcmp(subsession->codecName(), "L16") == 0))
			{
				// Used for ID
				int nMediaCount = m_vLiveMediaSinks.size();
				LiveDirectShowSink* pSink = WavLiveDirectShowSink::createNew(nMediaCount, *env, fileSinkBufferSize, pWatchVar);
				pSink->setMediaQueue(m_pSourceFilter->GetOutputPin(nMediaCount));
				subsession->sink = pSink;
				m_vLiveMediaSinks.push_back(pSink);
			} else {
				// Normal case:
				bool oneFilePerFrame = true;
				FileSink* fileSink = FileSink::createNew(*env, outFileName, 1000, oneFilePerFrame);
				subsession->sink = fileSink;
			}
			//////////////////////////////////////////////////////////////////////////
			
			//if (strcmp(subsession->mediumName(), "audio") == 0 &&
			//	(strcmp(subsession->codecName(), "AMR") == 0 ||
			//	strcmp(subsession->codecName(), "AMR-WB") == 0)) {
			//		// For AMR audio streams, we use a special sink that inserts AMR frame hdrs:
			//		fileSink = AMRAudioFileSink::createNew(*env, outFileName,
			//			fileSinkBufferSize, oneFilePerFrame);
			//} else if (strcmp(subsession->mediumName(), "video") == 0 &&
			//	(strcmp(subsession->codecName(), "H264") == 0)) {
			//		// For H.264 video stream, we use a special sink that insert start_codes:
			//		fileSink = H264VideoFileSink::createNew(*env, outFileName,
			//			fileSinkBufferSize, oneFilePerFrame);
			//} else {
			//	// Normal case:
			//	fileSink = FileSink::createNew(*env, outFileName,
			//		fileSinkBufferSize, oneFilePerFrame);
			//}
			//subsession->sink = fileSink;
			if (subsession->sink == NULL) {
				*env << "Failed to create FileSink for \"" << outFileName
					<< "\": " << env->getResultMsg() << "\n";
			} else 
			{
				if (strcmp(subsession->mediumName(), "video") == 0 &&
					strcmp(subsession->codecName(), "MP4V-ES") == 0 &&
					subsession->fmtp_config() != NULL) {
						// For MPEG-4 video RTP streams, the 'config' information
						// from the SDP description contains useful VOL etc. headers.
						// Insert this data at the front of the output file:
						unsigned configLen;
						unsigned char* configData
							= parseGeneralConfigStr(subsession->fmtp_config(), configLen);
						struct timeval timeNow;
						gettimeofday(&timeNow, NULL);
						fileSink->addData(configData, configLen, timeNow);
						delete[] configData;
				}

				subsession->sink->startPlaying(*(subsession->readSource()),
					NULL/*subsessionAfterPlaying*/,
					NULL/*subsession*/);

				// Also set a handler to be called if a RTCP "BYE" arrives
				// for this subsession:
				if (subsession->rtcpInstance() != NULL) 
				{
					subsession->rtcpInstance()->setByeHandler(&RtspClientSession::subsessionByeHandler, this/*subsession*/);
				}

				madeProgress = True;
			}
		}
		return madeProgress;
	}
}

bool RtspClientSession::startPlayingStreams()
{
	if (!m_pRtspClient->playMediaSession(*m_pSession)) 
	{
		*env << "Failed to start playing session: " << env->getResultMsg() << "\n";
		return false;
	} else {
		*env << "Started playing session\n";
	}

	int qosMeasurementIntervalMS = 1000;
	if (qosMeasurementIntervalMS > 0) {
		// Begin periodic QOS measurements:
		beginQOSMeasurement();
	}

	// Watch for incoming packets (if desired):
	//TODO???
	//checkForPacketArrival(NULL);
	//TODO???
	//checkInterPacketGaps(NULL);

	return true;
}

void RtspClientSession::beginQOSMeasurement()
{
	// Set up a measurement record for each active subsession:
	struct timeval startTime;
	gettimeofday(&startTime, NULL);
	nextQOSMeasurementUSecs = startTime.tv_sec*1000000 + startTime.tv_usec;
	qosMeasurementRecord* qosRecordTail = NULL;
	MediaSubsessionIterator iter(*m_pSession);
	MediaSubsession* subsession;
	while ((subsession = iter.next()) != NULL) {
		RTPSource* src = subsession->rtpSource();

		if (src == NULL) continue;

		qosMeasurementRecord* qosRecord
			= new qosMeasurementRecord(startTime, src);
		if (qosRecordHead == NULL) qosRecordHead = qosRecord;
		if (qosRecordTail != NULL) qosRecordTail->fNext = qosRecord;
		qosRecordTail  = qosRecord;
	}

	// Then schedule the first of the periodic measurements:
	scheduleNextQOSMeasurement();
}

void RtspClientSession::scheduleNextQOSMeasurement()
{
	nextQOSMeasurementUSecs += qosMeasurementIntervalMS*1000;
	struct timeval timeNow;
	gettimeofday(&timeNow, NULL);
	unsigned timeNowUSecs = timeNow.tv_sec*1000000 + timeNow.tv_usec;
	unsigned usecsToDelay = nextQOSMeasurementUSecs < timeNowUSecs ? 0
		: nextQOSMeasurementUSecs - timeNowUSecs;

	qosMeasurementTimerTask = env->taskScheduler().scheduleDelayedTask(
		usecsToDelay, (TaskFunc*)RtspClientSession::doPeriodicQOSMeasurement, (void*)this);
}

void RtspClientSession::doPeriodicQOSMeasurement(void* clientData) 
{
	RtspClientSession* pRtspClientSession = (RtspClientSession*)clientData;
	pRtspClientSession->periodicQOSMeasurement();
}

void RtspClientSession::periodicQOSMeasurement() 
{
	struct timeval timeNow;
	gettimeofday(&timeNow, NULL);

	for (qosMeasurementRecord* qosRecord = qosRecordHead;
		qosRecord != NULL; qosRecord = qosRecord->fNext) {
			qosRecord->periodicQOSMeasurement(timeNow);
	}

	// Do this again later:
	scheduleNextQOSMeasurement();
}

bool RtspClientSession::initialiseRtspSession( const char* szUrl )
{
	// Get options response
	char* optionsResponse = m_pRtspClient->sendOptionsCmd(szUrl, NULL, NULL, NULL, m_nTimeoutS);
	*env << "RTSP Options response: " << optionsResponse << "\r\n";
	delete[] optionsResponse;

	// Get SDP description
	char* sdpDescription = m_pRtspClient->describeURL(szUrl, NULL, false, m_nTimeoutS);
	if (sdpDescription == NULL) 
	{
		m_sLastError = "Failed to get a SDP description from URL \"" + std::string(szUrl) + "\": " + std::string(env->getResultMsg()) + "\n";
		*env << "Failed to get a SDP description from URL \"" << szUrl << "\": " << env->getResultMsg() << "\n";
		return false;
	}

	*env << "Opened URL \"" << szUrl << "\", returning a SDP description:\n" << sdpDescription << "\n";
	m_statusCode = m_pRtspClient->describeStatus();

	// Create a media session object from this SDP description:
	m_pSession = MediaSession::createNew(*env, sdpDescription);
	delete[] sdpDescription;
	if (m_pSession == NULL) 
	{
		m_sLastError = "Failed to create a MediaSession object from the SDP description: " + std::string(env->getResultMsg()) + "\n";
		*env << "Failed to create a MediaSession object from the SDP description: " << env->getResultMsg() << "\n";
		return false;
	} 
	else if (!m_pSession->hasSubsessions()) 
	{
		m_sLastError = "This session has no media subsessions (i.e., \"m=\" lines)\n";
		*env << "This session has no media subsessions (i.e., \"m=\" lines)\n";
		return false;
	}

	return true;
}

bool RtspClientSession::setupAndPlayRtspSession()
{
	// Create RTP sources	
	if (!createRtpSources())
	{
		m_sLastError = "Failed to create RTP sources";
		return false;
	}

	// Setup RTP sources	
	if (!setupStreams())
	{
		m_sLastError = "Failed to setup RTP streams";
		return false;
	}

	// Create RTP Sinks
	if (!createRtpReceivers())
	{
		m_sLastError = "Failed to create RTP receivers";
		return false;
	}

	// Play streams
	if (!startPlayingStreams())
	{
		m_sLastError = "Failed to start playing streams";
		return false;
	}

	// Eventloop
	*env << "Starting the live555 event loop for playing the data\n";
	env->taskScheduler().doEventLoop(m_pSourceFilter->GetRtspWatchVar()); // does not return	

	return true;
}

void RtspClientSession::qosMeasurementRecord::periodicQOSMeasurement(struct timeval const& timeNow) 
{
	unsigned secsDiff = timeNow.tv_sec - measurementEndTime.tv_sec;
	int usecsDiff = timeNow.tv_usec - measurementEndTime.tv_usec;
	double timeDiff = secsDiff + usecsDiff/1000000.0;
	measurementEndTime = timeNow;

	RTPReceptionStatsDB::Iterator statsIter(fSource->receptionStatsDB());
	// Assume that there's only one SSRC source (usually the case):
	RTPReceptionStats* stats = statsIter.next(True);
	if (stats != NULL) {
		double kBytesTotalNow = stats->totNumKBytesReceived();
		double kBytesDeltaNow = kBytesTotalNow - kBytesTotal;
		kBytesTotal = kBytesTotalNow;

		double kbpsNow = timeDiff == 0.0 ? 0.0 : 8*kBytesDeltaNow/timeDiff;
		if (kbpsNow < 0.0) kbpsNow = 0.0; // in case of roundoff error
		if (kbpsNow < kbits_per_second_min) kbits_per_second_min = kbpsNow;
		if (kbpsNow > kbits_per_second_max) kbits_per_second_max = kbpsNow;

		unsigned totReceivedNow = stats->totNumPacketsReceived();
		unsigned totExpectedNow = stats->totNumPacketsExpected();
		unsigned deltaReceivedNow = totReceivedNow - totNumPacketsReceived;
		unsigned deltaExpectedNow = totExpectedNow - totNumPacketsExpected;
		totNumPacketsReceived = totReceivedNow;
		totNumPacketsExpected = totExpectedNow;

		double lossFractionNow = deltaExpectedNow == 0 ? 0.0
			: 1.0 - deltaReceivedNow/(double)deltaExpectedNow;
		//if (lossFractionNow < 0.0) lossFractionNow = 0.0; //reordering can cause
		if (lossFractionNow < packet_loss_fraction_min) {
			packet_loss_fraction_min = lossFractionNow;
		}
		if (lossFractionNow > packet_loss_fraction_max) {
			packet_loss_fraction_max = lossFractionNow;
		}
	}
}