/** RtspDataSession.cpp
===========================================================================
MODULE          : RtspDataSession: 

REVISION HISTORY: Version 1.0

COPYRIGHT       : (c)CSIR, Meraka Institute 2008 all rights reserved

RESTRICTIONS	: The information/data/code contained within this file is 
				  the property of CSIR, Meraka Institute and has been 
				  classified as PUBLIC.
===========================================================================
*/

#include "RtspDataSession.h"

// LiveMedia
#include "BasicUsageEnvironment.hh"

// RTVC
#include "RtspHelper.h"
#include "RtvcRtpSink.h"
#include "RtpPacketManager.h"

RtspDataSession::RtspDataSession(RtpPacketManager* pRtpPacketManager)
:m_pRtpPacketManager(pRtpPacketManager),
	m_pRtspClient(NULL),
	m_pSession(NULL),
	m_bStreamUsingTCP(true),
	m_watchVariable(0),
	m_sLastError("")
{;}

bool RtspDataSession::startRetrievingData(const std::string sUrl, int nTimeOut)
{
	const char* szUrl = sUrl.c_str();

	// Setup environment
	TaskScheduler* pScheduler = BasicTaskScheduler::createNew();
	UsageEnvironment* pEnv = BasicUsageEnvironment::createNew(*pScheduler);

	bool bSuccess = true;

	do
	{
		// Create client
		m_pRtspClient = RTSPClient::createNew(*pEnv);

		if (!RtspHelper::createMediaSession(sUrl, m_pRtspClient, &m_pSession, m_sLastError))
		{
			bSuccess = false;
			break;
		}
		
		bSuccess = playStreams(pEnv, m_pSession);
		break;
	}while (true);

	// Cleanup RTSP Media session
	RtspHelper::shutdownMediaSession(m_pRtspClient, m_pSession);
	return bSuccess;
}

void RtspDataSession::shutdown()
{
	// Cleanup RTSP Media session
	RtspHelper::shutdownMediaSession(m_pRtspClient, m_pSession);
}

bool RtspDataSession::playStreams(UsageEnvironment* pEnv, MediaSession* pSession)
{
	// Setup and play the streams 
	if (!RtspHelper::createRtpSources(pEnv, pSession))
	{
		m_sLastError = "Failed to create RTP sources";
		return false;
	}

	// Setup RTP source
	if (!RtspHelper::setupStreams(m_pRtspClient, pSession, m_bStreamUsingTCP))
	{
		m_sLastError = "Failed to setup RTP streams";
		return false;
	}

	if (!createReceivers(pSession))
	{
		m_sLastError =  "Failed to create RTP receivers: " +  std::string(pEnv->getResultMsg());
		return false;
	}

	// Play streams
	if (!m_pRtspClient->playMediaSession(*pSession)) 
	{
		m_sLastError =  "Failed to start playing session: " +  std::string(pEnv->getResultMsg());
		return false;
	}

	// Start the liveMedia eventloop
	pEnv->taskScheduler().doEventLoop(&m_watchVariable); // does not return	until watch variable is set
	return true;
}

bool RtspDataSession::createReceivers(MediaSession* pSession)
{
	// Create and start media sinks for each subsession:
	bool madeProgress = False;
	MediaSubsessionIterator iter(*m_pSession);
	MediaSubsession *pSubsession;
	int nSinkID = -1;
	while ((pSubsession = iter.next()) != NULL) 
	{
		++nSinkID;

		// Create our RTPSinks that will receive the data
		if (pSubsession->readSource() == NULL) continue;// was not initiated

		int nSinkBufferSize = 20000;

		// Create a new queue in the packet manager
		m_pRtpPacketManager->createQueue(nSinkID);
		// Link the Rtp sink to the ID of the queue in the packet manager
		RtvcRtpSink<RtpPacketManager>* pSink = new RtvcRtpSink<RtpPacketManager>(pSession->envir(), nSinkID, nSinkBufferSize, m_pRtpPacketManager);
		pSubsession->sink = pSink;

		// TODO Set after playing pointers
		pSubsession->sink->startPlaying(*(pSubsession->readSource()), NULL, NULL);

		// Also set a handler to be called if a RTCP "BYE" arrives
		// for this subsession:
		if (pSubsession->rtcpInstance() != NULL) 
		{
			pSubsession->rtcpInstance()->setByeHandler(&RtspDataSession::subsessionByeHandler, this/*subsession*/);
		}
		madeProgress = True;
	}
	return madeProgress;
}

