/** RtvcRtpSink.h
===========================================================================
MODULE          : RtvcRtpSink

DESCRIPTIONS    : LiveMedia extension - Base RTP Sink for received media. This class is the interface between the liveMedia RTP library 
				  and our DirectShow media queue. It receive raw media data from the RTP Source and adds it to the media queue. 

REVISION HISTORY: Version 1.0

COPYRIGHT       : (c)CSIR, Meraka Institute 2008 all rights reserved

RESTRICTIONS	: The information/data/code contained within this file is 
				  the property of CSIR, Meraka Institute and has been 
				  classified as PUBLIC.
===========================================================================
*/
#pragma once

// DirectShow includes
#pragma warning(push)     // disable for this header only
#pragma warning(disable:4312) 
// DirectShow
#include <Streams.h>
#pragma warning(pop)      // restore original warning level

#include <Windows.h>

// STL
#include <iostream>

// LiveMedia
#include "MediaSink.hh"
#include <RTPSource.hh>

// RTVC
#include <Shared/MediaSample.h>


template<class T>
class RtvcRtpSink : public MediaSink
{
public:
	/// Constructor
	RtvcRtpSink(UsageEnvironment& env, int nSourceID, unsigned bufferSize, T* pSampleHandler)
		: MediaSink(env),
		m_pSampleHandler(pSampleHandler),
		m_bHasBeenSyncedUsingRtcp(false),
		m_nSourceID(nSourceID)
	{
		// Create memory buffer
		fBuffer = new unsigned char[bufferSize];
		fBufferSize = bufferSize;
	}

	/// Destructor
	virtual ~RtvcRtpSink( void )
	{
		// Free memory
		delete[] fBuffer;
	}
	
protected:
	/// LiveMedia method
	static void afterGettingFrame( void* clientData, unsigned frameSize, unsigned numTruncatedBytes, struct timeval presentationTime, unsigned durationInMicroseconds )
	{
		RtvcRtpSink* sink = (RtvcRtpSink*)clientData;
		sink->afterGettingFrame1(frameSize, presentationTime);
	}

	/// LiveMedia method
	virtual void afterGettingFrame1( unsigned frameSize, struct timeval presentationTime )
	{
		processData(fBuffer, frameSize, presentationTime);
		// Then try getting the next frame:
		continuePlaying();
	}

	/// LiveMedia method
	virtual Boolean continuePlaying()
	{
		if (fSource == NULL) return False;

		fSource->getNextFrame(fBuffer, fBufferSize,
			afterGettingFrame, this,
			onSourceClosure, this);

		return True;
	}

	void processData( unsigned char* data, unsigned dataSize, struct timeval presentationTime )
	{
		// Check if stream has been synced using RTCP: to do this we need to get hold of the RTPSource object
		RTPSource* pRtpSource = (RTPSource*)fSource;
		// Convert presentation time to double
		double dStartTime = presentationTime.tv_sec + (presentationTime.tv_usec / 1000000.0);
		// Implicit Template Declaration
		m_pSampleHandler->processMediaSample(m_nSourceID, data, dataSize, dStartTime, pRtpSource->hasBeenSynchronizedUsingRTCP());
	}

private:
	/// The handler that processes the media samples
	T* m_pSampleHandler;

	/// Buffer for the media sample raw data
	unsigned char* fBuffer;
	/// Buffer size
	unsigned fBufferSize;

	/// This member gets set when the first RTCP synced sample is received
	bool m_bHasBeenSyncedUsingRtcp;

	/// ID of the sink which is the "source" of the output pin
	int m_nSourceID;
};
