/** @file
===========================================================================
MODULE				: CCustomBaseFilter

DESCRIPTIONS		: Base class for Meraka Custom Filters

REVISION HISTORY	: Version 1.0

COPYRIGHT			: (c)CSIR, Meraka Institute 2007 all rights reserved

RESTRICTIONS		: The information/data/code contained within this file is 
					  the property of CSIR, Meraka Institute and has been 
					  classified as PUBLIC.
===========================================================================
*/

#pragma once
#include <DirectShow/stdafx.h>
#include <DirectShow/CStatusInterface.h>
#include <DirectShow/CSettingsInterface.h>

using namespace std;

class CCustomBaseFilter : public CTransformFilter, public CSettingsInterface, public CStatusInterface
{
	///this needs to be declared for the extra interface (adds the COM AddRef, etc methods)
	DECLARE_IUNKNOWN;

public:
	CCustomBaseFilter(TCHAR *pObjectName, LPUNKNOWN lpUnk, CLSID clsid);
	virtual ~CCustomBaseFilter(void);
	
	/// Overridden method which validates the input media type against the ones accepted by this filter
	virtual HRESULT CheckInputType(const CMediaType *pMediaType);

	/// Helper method which allows us to capture the input output media type properties
	/// Override this method to set other initialization code but call this base class method first
	virtual HRESULT SetMediaType(PIN_DIRECTION direction, const CMediaType *pmt);

	/// Transform code
	/**
	* This method takes the source sample and transforms it and writes the transformed sample to the destination
	* This method is called by the transform input pin's Receive method.
	* @param pSource The source sample
	* @param pDest The destination sample
	*/
	HRESULT Transform(IMediaSample *pSource, IMediaSample *pDest);

	/// override this to publicize our interfaces
	STDMETHODIMP NonDelegatingQueryInterface(REFIID riid, void **ppv);
	
	///Overridden from ISettingsInterface
	STDMETHODIMP GetParameter( const char* szParamName, char* szResult, int& length );
	STDMETHODIMP SetParameter( const char* szParamName, const char* szValue);

	/// This method needs to overridden by child classes - add all possible accepted input combinations
	/// using the AddInputType method
	virtual void InitialiseInputTypes() PURE;

protected:
	/**
	* Used for Media Type Negotiation 
	* Call the AddInputType method in your overridden InitialiseInputTypes 
	* method to add your acceptable media types, subtypes and formats.
	*/	
	void AddInputType(const GUID* pType, const GUID* pSubType, const GUID* pFormat);

	/**
	* This method must be overridden in the sub class to apply the filter specific transformation
	* It must return the size of the resulting transformed output buffer
	*/
	virtual DWORD ApplyTransform(BYTE* pBufferIn, BYTE* pBufferOut) PURE;

	/// Video header of input
	VIDEOINFOHEADER m_videoInHeader;
	/// Video header of output
	VIDEOINFOHEADER m_videoOutHeader;
	// Other video information
	/// Height of image
	int m_nInHeight;
	/// Width of image
	int m_nInWidth;
	/// Number of pixels in image
	int m_nInPixels;
	/// Height in output image
	int m_nOutHeight;
	/// Width in output image
	int m_nOutWidth;
	/// Number of pixels in output image
	int m_nOutPixels;
	/// Bit count of image
	int m_nBitCount;

private:
	// Used for storing and checking connection types
	GUID_LIST m_vInTypes;
	GUID_LIST m_vInSubTypes;
	GUID_LIST m_vInFormats;
};
