/** @file
===========================================================================
MODULE				: MultiIOBaseFilter

DESCRIPTIONS		: This base class takes care of doing most of the menial work when writing a transform filter that has multiple input and output pins.

REVISION HISTORY	: Version 1.0

COPYRIGHT			: (c)CSIR, Meraka Institute 2007 all rights reserved

RESTRICTIONS		: The information/data/code contained within this file is 
					  the property of CSIR, Meraka Institute and has been 
					  classified as PUBLIC.
===========================================================================
*/

#include "MultiIOInputPin.h"

#include "MultiIOBaseFilter.h"

CMultiIOInputPin::CMultiIOInputPin(CMultiIOBaseFilter* pFilter, CCritSec* pLock, HRESULT* phr, LPCWSTR pName, int nIndex)
: CBaseInputPin(NAME("TM Multi Input Pin"), pFilter, pLock, phr, pName),
m_pFilter(pFilter),
m_nIndex(nIndex)
{
}

CMultiIOInputPin::~CMultiIOInputPin(void)
{
}

HRESULT CMultiIOInputPin::CheckMediaType( const CMediaType* mtIn )
{
	return m_pFilter->CheckInputType(mtIn);
}

HRESULT CMultiIOInputPin::BreakConnect()
{
	HRESULT hr = CBaseInputPin::BreakConnect();
	m_pFilter->OnDisconnect(m_nIndex);
	return hr;
}

HRESULT CMultiIOInputPin::CompleteConnect( IPin *pReceivePin )
{
	HRESULT hr = CBaseInputPin::CompleteConnect(pReceivePin);
	if (SUCCEEDED(hr))
	{
		m_pFilter->OnConnect(m_nIndex, RTVC_IN);
	}
	return hr;
}

STDMETHODIMP CMultiIOInputPin::Receive( IMediaSample *pSample )
{
	//Copied from CTransformInputPin
	HRESULT hr;
	CAutoLock lck(&m_pFilter->m_csReceive);
	ASSERT(pSample);

	// check all is well with the base class
	hr = CBaseInputPin::Receive(pSample);
	if (S_OK == hr) {
		hr = m_pFilter->Receive(pSample, m_nIndex);
	}
	return hr;
}

HRESULT CMultiIOInputPin::SetMediaType( const CMediaType* mtIn )
{
	// Set the base class media type (should always succeed)
	HRESULT hr = CBasePin::SetMediaType(mtIn);
	if (FAILED(hr)) {
		return hr;
	}

	// check the transform can be done (should always succeed)
	ASSERT(SUCCEEDED(m_pFilter->CheckInputType(mtIn)));

	return m_pFilter->SetMediaType(PINDIR_INPUT, mtIn, m_nIndex);
}

STDMETHODIMP CMultiIOInputPin::EndOfStream()
{
	HRESULT hr = CBaseInputPin::EndOfStream();
	if (FAILED(hr))
	{
		return hr;
	}
	return m_pFilter->EndOfStream(m_nIndex);
}

STDMETHODIMP CMultiIOInputPin::BeginFlush()
{
	HRESULT hr = CBaseInputPin::BeginFlush();
	if (FAILED(hr))
	{
		return hr;
	}
	return m_pFilter->BeginFlush(m_nIndex);
}

STDMETHODIMP CMultiIOInputPin::EndFlush()
{
	HRESULT hr = CBaseInputPin::EndFlush();
	if (FAILED(hr))
	{
		return hr;
	}
	return m_pFilter->EndFlush(m_nIndex);
}

STDMETHODIMP CMultiIOInputPin::NewSegment( REFERENCE_TIME tStart, REFERENCE_TIME tStop, double dRate )
{
	HRESULT hr = CBaseInputPin::NewSegment(tStart, tStop, dRate);
	if (FAILED(hr))
	{
		return hr;
	}
	return m_pFilter->NewSegment(tStart, tStop, dRate, m_nIndex );
}